SystemOrganization addCategory: #ObjectUsageTracer!

!Class methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 12/30/2011 12:23'!
hasUsedInstances 
	^ self allInstances anySatisfy: [:inst | ObjectUsageTracer new isMarked: inst ]. 
	! !

!Class methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 12/30/2011 12:24'!
unusedInstances
	^ self allInstances reject: [:inst | ObjectUsageTracer new isMarked: inst  ]. 
	! !

!Class methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 12/30/2011 12:25'!
unusedMethods
	^ self methods reject: [:method |  ObjectUsageTracer new isMarked: method ]. 
	! !

!Class methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 12/30/2011 12:26'!
usedInstances
	^ self allInstances select: [:inst | ObjectUsageTracer new isMarked: inst ]. 
	! !

!Class methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 12/30/2011 12:26'!
usedMethods
	^ self methods select: [:method | ObjectUsageTracer new isMarked: method ]. 
	! !

!Object methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 5/31/2012 15:26'!
isUsed
	^ self experimentalBit! !

Object subclass: #ObjectUsageTracer
	instanceVariableNames: ''
	classVariableNames: 'HasToTraceByDefault'
	poolDictionaries: ''
	category: 'ObjectUsageTracer'!

!ObjectUsageTracer commentStamp: '<historical>' prior: 0!
This class lets you find used objects during an execution. In order to do that:

1) You first have to evauate: ''ObjectUsageTracer new startTracing.''
With that, you mark all objects as unused so that you can start a clean discovery.

2) Then, you do things. Run a seaside app, develope, code, etc. 

3) Once you are finished and want to see the usage, print or inspect: ''ObjectUsageTracer new stopTracingAndGetStadistics''
This will get the statistics and unmark all objects again.

Another option is to do: ''ObjectUsageTracer trace: aBlock''
where aBlock is what you want to analyze.

Example:

ObjectUsageTracer new startTracing.
Transcript show: 'mariano'.
"....."
ObjectUsageTracer new stopTracingAndGetStadistics.
!

!ObjectUsageTracer class methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/29/2011 20:23'!
disableTracingByDefault
	"This will cause that when the image starts, the system does NOT start tracing objects by default. It will be like that until someone sends #enableTracing."
	HasToTraceByDefault := false.
	Smalltalk recreateSpecialObjectsArray.! !

!ObjectUsageTracer class methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/29/2011 20:23'!
enableTracingByDefault
	"This will cause that when the image starts, the system starts tracing objects by default. It will be like that until someone sends #disableTracing."
	HasToTraceByDefault := true.
	Smalltalk recreateSpecialObjectsArray.! !

!ObjectUsageTracer class methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/29/2011 19:53'!
hasToTraceByDefault
	HasToTraceByDefault isNil ifTrue: [ HasToTraceByDefault := false ].
	^ HasToTraceByDefault! !

!ObjectUsageTracer methodsFor: 'enable - disable' stamp: 'MarianoMartinezPeck 12/22/2011 12:16'!
disableTracing
	self primitiveSetHasToTrace: false.
! !

!ObjectUsageTracer methodsFor: 'enable - disable' stamp: 'MarianoMartinezPeck 12/22/2011 11:34'!
enableTracing
	self primitiveSetHasToTrace: true.
! !

!ObjectUsageTracer methodsFor: 'helpers' stamp: 'LucFabresse 12/25/2011 20:04'!
getStatistics
	| anArray stadistics|
	anArray := self primitiveGetStatistics.
	stadistics := ObjectUsageTracerStatistics fromPrimitiveArray: anArray.
	^ stadistics ! !

!ObjectUsageTracer methodsFor: 'marking' stamp: 'MarianoMartinezPeck 12/22/2011 11:54'!
isMarked: anObject
	^ anObject isExperimentalBitOn! !

!ObjectUsageTracer methodsFor: 'marking' stamp: 'MarianoMartinezPeck 12/22/2011 11:53'!
markObject: anObject
	anObject experimentalBit: true.! !

!ObjectUsageTracer methodsFor: 'primitives' stamp: 'LucFabresse 12/25/2011 20:03'!
primitiveGetStatistics 

	<primitive: 154>	"successful completion returns an Array with the stadistics numbers"
	^  self primitiveFailed 
! !

!ObjectUsageTracer methodsFor: 'primitives' stamp: 'MarianoMartinezPeck 12/22/2011 11:33'!
primitiveSetHasToTrace: aBoolean

	<primitive: 153>	"successful completion returns self"
	^  self primitiveFailed 
! !

!ObjectUsageTracer methodsFor: 'start - stop' stamp: 'MarianoMartinezPeck 12/22/2011 11:36'!
startTracing
	self unmarkAllObjects.
	self enableTracing.
	! !

!ObjectUsageTracer methodsFor: 'start - stop' stamp: 'MarianoMartinezPeck 12/22/2011 11:36'!
stopTracing
	self disableTracing.
	! !

!ObjectUsageTracer methodsFor: 'start - stop' stamp: 'LucFabresse 12/25/2011 20:04'!
stopTracingAndGetStadistics
	self disableTracing.
	^ self getStatistics.
! !

!ObjectUsageTracer methodsFor: 'tracing' stamp: 'MarianoMartinezPeck 12/22/2011 12:41'!
trace: aBlock
	self startTracing.
	aBlock value.
	self stopTracing.! !

!ObjectUsageTracer methodsFor: 'tracing' stamp: 'MarianoMartinezPeck 12/29/2011 14:48'!
traceAndGetStatistics: aBlock
	self trace: aBlock.
	^ self getStatistics.! !

!ObjectUsageTracer methodsFor: 'tracing' stamp: 'MarianoMartinezPeck 12/27/2011 11:28'!
traceCleaningCache: aBlock
	Object flushCache.
	self startTracing.
	aBlock value.
	self stopTracing.! !

!ObjectUsageTracer methodsFor: 'helpers' stamp: 'MarianoMartinezPeck 12/22/2011 11:35'!
unmarkAllObjects
	ExperimentalBitHandler turnOffExperimentalBitOfAllObjects! !

!ObjectUsageTracer methodsFor: 'marking' stamp: 'MarianoMartinezPeck 12/22/2011 12:17'!
unmarkObject: anObject
	anObject experimentalBit: false.! !

Object subclass: #ObjectUsageTracerStatistics
	instanceVariableNames: 'unusedCount usedCount objectCount unusedPercentage usedPercentage usedMemory unusedMemory totalMemory percentageUsedMemory percentageUnusedMemory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ObjectUsageTracer'!

!ObjectUsageTracerStatistics class methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
fromPrimitiveArray: anArray
	| usedCount unusedCount usedMemory unusedMemory objectsCount totalMemory |
	usedCount := (anArray at: 1).
	unusedCount := (anArray at: 2).
	usedMemory  := (anArray at: 3).
	unusedMemory  := (anArray at: 4).
	
	objectsCount := usedCount + unusedCount.
	totalMemory := unusedMemory + usedMemory.
	^ self new
		usedCount: usedCount;
		unusedCount: unusedCount;
		objectCount: objectsCount;
		unusedPercentage: ((unusedCount * 100) / objectsCount) asFloat;
		usedPercentage: ((usedCount * 100) / objectsCount) asFloat;
		usedMemory: usedMemory;
		unusedMemory: unusedMemory;
		totalMemory: totalMemory;
		percentageUnusedMemory: ((unusedMemory * 100) / totalMemory) asFloat;
		percentageUsedMemory: ((usedMemory * 100) / totalMemory) asFloat
		 ! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
objectCount
	^ objectCount! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
objectCount: anObject
	objectCount := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
percentageUnusedMemory
	^ percentageUnusedMemory! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
percentageUnusedMemory: anObject
	percentageUnusedMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
percentageUsedMemory
	^ percentageUsedMemory! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
percentageUsedMemory: anObject
	percentageUsedMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
printOn: aStream
	aStream 
		nextPutAll: 'Total amount of objects: ', self objectCount asString, String cr;
		nextPutAll: 'Amount of used objects: ', self usedCount asString, String cr;
		nextPutAll: 'Percentage of used objects: ', self usedPercentage asString, String cr;
		nextPutAll: 'Amount of unused objects: ', self unusedCount asString, String cr;
		nextPutAll: 'Percentage of unused objects: ', self unusedPercentage asString, String cr;
		nextPutAll: String cr;
		nextPutAll: 'Total bytes of memory used: ', self totalMemory  asString, String cr;
		nextPutAll: 'Bytes of memory for the used objects: ', self usedMemory asString, String cr;
		nextPutAll: 'Percentage of memory for the used objects: ', self percentageUsedMemory asString, String cr;
		nextPutAll: 'Bytes of memory for the unused objects: ', self unusedMemory asString, String cr;
		nextPutAll: 'Percentage of memory for the unused objects: ', self percentageUnusedMemory asString, String cr.
	! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
totalMemory
	^ totalMemory! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
totalMemory: anObject
	totalMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedCount
	^ unusedCount! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedCount: anObject
	unusedCount := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedMemory
	^ unusedMemory! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedMemory: anObject
	unusedMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedPercentage
	^ unusedPercentage! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedPercentage: anObject
	unusedPercentage := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedCount
	^ usedCount! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedCount: anObject
	usedCount := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedMemory
	^ usedMemory ! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedMemory: anObject
	usedMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedPercentage
	^ usedPercentage! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedPercentage: anObject
	usedPercentage := anObject! !

!SmalltalkImage methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 12/29/2011 19:51'!
recreateSpecialObjectsArray
	"Smalltalk recreateSpecialObjectsArray"
	
	"To external package developers:
	**** DO NOT OVERRIDE THIS METHOD.  *****
	If you are writing a plugin and need additional special object(s) for your own use, 
	use addGCRoot() function and use own, separate special objects registry "
	
	"The Special Objects Array is an array of objects used by the Squeak virtual machine.
	 Its contents are critical and accesses to it by the VM are unchecked, so don't even
	 think of playing here unless you know what you are doing."
	| newArray |
	newArray := Array new: 59.
	"Nil false and true get used throughout the interpreter"
	newArray at: 1 put: nil.
	newArray at: 2 put: false.
	newArray at: 3 put: true.
	"This association holds the active process (a ProcessScheduler)"
	newArray at: 4 put: (self associationAt: #Processor).
	"Numerous classes below used for type checking and instantiation"
	newArray at: 5 put: Bitmap.
	newArray at: 6 put: SmallInteger.
	newArray at: 7 put: ByteString.
	newArray at: 8 put: Array.
	newArray at: 9 put: Smalltalk.
	newArray at: 10 put: Float.
	newArray at: 11 put: MethodContext.
	newArray at: 12 put: BlockContext.
	newArray at: 13 put: Point.
	newArray at: 14 put: LargePositiveInteger.
	newArray at: 15 put: Display.
	newArray at: 16 put: Message.
	newArray at: 17 put: CompiledMethod.
	newArray at: 18 put: (self specialObjectsArray at: 18).
	"(low space Semaphore)"
	newArray at: 19 put: Semaphore.
	newArray at: 20 put: Character.
	newArray at: 21 put: #doesNotUnderstand:.
	newArray at: 22 put: #cannotReturn:.
	newArray at: 23 put: nil. "This is the process signalling low space."
	"An array of the 32 selectors that are compiled as special bytecodes,
	 paired alternately with the number of arguments each takes."
	newArray at: 24 put: #(	#+ 1 #- 1 #< 1 #> 1 #<= 1 #>= 1 #= 1 #~= 1
							#* 1 #/ 1 #\\ 1 #@ 1 #bitShift: 1 #// 1 #bitAnd: 1 #bitOr: 1
							#at: 1 #at:put: 2 #size 0 #next 0 #nextPut: 1 #atEnd 0 #== 1 #class 0
							#blockCopy: 1 #value 0 #value: 1 #do: 1 #new 0 #new: 1 #x 0 #y 0 ).
	"An array of the 255 Characters in ascii order.
	 Cog inlines table into machine code at: prim so do not regenerate it."
	newArray at: 25 put: (self specialObjectsArray at: 25).
	newArray at: 26 put: #mustBeBoolean.
	newArray at: 27 put: ByteArray.
	newArray at: 28 put: Process.
	"An array of up to 31 classes whose instances will have compact headers"
	newArray at: 29 put: self compactClassesArray.
	newArray at: 30 put: (self specialObjectsArray at: 30).
	"(delay Semaphore)"
	newArray at: 31 put: (self specialObjectsArray at: 31).
	"(user interrupt Semaphore)"
	"Prototype instances that can be copied for fast initialization"
	newArray at: 32 put: (Float new: 2).
	newArray at: 33 put: (LargePositiveInteger new: 4).
	newArray at: 34 put: 0@0.
	newArray at: 35 put: #cannotInterpret:.
	"Note: This must be fixed once we start using context prototypes (yeah, right)"
	"(MethodContext new: CompiledMethod fullFrameSize)."
	newArray at: 36 put: (self specialObjectsArray at: 36). "Is the prototype MethodContext (unused by the VM)"
	newArray at: 37 put: BlockClosure.
	"(BlockContext new: CompiledMethod fullFrameSize)."
	newArray at: 38 put: (self specialObjectsArray at: 38). "Is the prototype BlockContext (unused by the VM)"
	"array of objects referred to by external code"
	newArray at: 39 put: (self specialObjectsArray at: 39).	"preserve external semaphores"
	newArray at: 40 put: nil. "Reserved for Mutex in Cog VMs"
	newArray at: 41 put: nil. "Reserved for a LinkedList instance for overlapped calls in CogMT"
	"finalization Semaphore"
	newArray at: 42 put: ((self specialObjectsArray at: 42) ifNil: [Semaphore new]).
	newArray at: 43 put: LargeNegativeInteger.
	"External objects for callout.
	 Note: Written so that one can actually completely remove the FFI."
	newArray at: 44 put: (self at: #ExternalAddress ifAbsent: []).
	newArray at: 45 put: (self at: #ExternalStructure ifAbsent: []).
	newArray at: 46 put: (self at: #ExternalData ifAbsent: []).
	newArray at: 47 put: (self at: #ExternalFunction ifAbsent: []).
	newArray at: 48 put: (self at: #ExternalLibrary ifAbsent: []).
	newArray at: 49 put: #aboutToReturn:through:.
	newArray at: 50 put: #run:with:in:.
	"51 reserved for immutability message"
	"newArray at: 51 put: #attemptToAssign:withIndex:."
	newArray at: 52 put: #(nil "nil => generic error" #'bad receiver'
							#'bad argument' #'bad index'
							#'bad number of arguments'
							#'inappropriate operation'  #'unsupported operation'
							#'no modification' #'insufficient object memory'
							#'insufficient C memory' #'not found' #'bad method'
							#'internal error in named primitive machinery'
							#'object may move').
	"53 to 55 are for Alien"
	newArray at: 53 put: (self at: #Alien ifAbsent: []).
	newArray at: 54 put: #invokeCallback:stack:registers:jmpbuf:.
	newArray at: 55 put: (self at: #UnsafeAlien ifAbsent: []).

	"Weak reference finalization"
	newArray at: 56 put: (self at: #WeakFinalizationList ifAbsent: []).
	
	"For enabling and disabling object usage tracing"
	newArray at: 59 put: ObjectUsageTracer hasToTraceByDefault.
	
	"Now replace the interpreter's reference in one atomic operation"
	self specialObjectsArray become: newArray
	! !
